use color_eyre::Result;
use crossterm::event::{KeyCode, KeyEvent};
use ratatui::{
    layout::{Constraint, Flex, Layout, Rect},
    Frame,
};
use tokio::sync::mpsc::UnboundedSender;

use super::Component;
use crate::{action::Action, app::Mode, config::Config};

mod latest_block;
mod finalized_block;
mod block_ticker;
mod current_era;
mod current_epoch;
mod extrinsics_chart;
mod block_explorer;
mod extrinsic_explorer;
mod log_explorer;

use latest_block::LatestBlock;
use finalized_block::FinalizedBlock;
use block_ticker::BlockTicker;
use current_era::CurrentEra;
use current_epoch::CurrentEpoch;
use extrinsics_chart::ExtrinsicsChart;
use block_explorer::BlockExplorer;
use extrinsic_explorer::ExtrinsicExplorer;
use log_explorer::LogExplorer;

#[derive(Debug, Clone, PartialEq)]
pub enum CurrentTab {
    Nothing,
    Blocks,
    Extrinsics,
}

pub trait PartialComponent: Component {
    fn set_active(&mut self, current_tab: CurrentTab);
}

pub struct Explorer {
    is_active: bool,
    current_tab: CurrentTab,
    components: Vec<Box<dyn PartialComponent>>
}

impl Default for Explorer {
    fn default() -> Self {
        Self {
            is_active: false,
            current_tab: CurrentTab::Nothing,
            components: vec![
                Box::new(BlockTicker::default()),
                Box::new(LatestBlock::default()),
                Box::new(FinalizedBlock::default()),
                Box::new(CurrentEra::default()),
                Box::new(CurrentEpoch::default()),
                Box::new(ExtrinsicsChart::default()),
                Box::new(BlockExplorer::default()),
                Box::new(ExtrinsicExplorer::default()),
                Box::new(LogExplorer::default()),
            ]
        }
    }
}

impl Explorer {
    fn move_left(&mut self) {
        if let CurrentTab::Extrinsics = self.current_tab {
            self.current_tab = CurrentTab::Blocks;
        }
    }

    fn move_right(&mut self) {
        match self.current_tab {
            CurrentTab::Nothing => self.current_tab = CurrentTab::Blocks,
            CurrentTab::Blocks => self.current_tab = CurrentTab::Extrinsics,
            _ => {}
        }
    }
}

impl PartialComponent for Explorer {
    fn set_active(&mut self, _current_tab: CurrentTab) {}
}

impl Component for Explorer {
    fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
        for component in self.components.iter_mut() {
            component.register_action_handler(tx.clone())?;
        }
        Ok(())
    }

    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(_) = config.styles.get(&crate::app::Mode::Explorer) {
            for component in self.components.iter_mut() {
                component.register_config_handler(config.clone())?;
            }
        }
        Ok(())
    }

    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        if !self.is_active { return Ok(None); }
        match key.code {
            KeyCode::Esc => {
                self.is_active = false;
                self.current_tab = CurrentTab::Nothing;
                for component in self.components.iter_mut() {
                    component.set_active(self.current_tab.clone());
                }
                return Ok(Some(Action::SetActiveScreen(Mode::Menu)));
            },
            KeyCode::Enter | KeyCode::Char('l') | KeyCode::Right => {
                self.move_right();
                for component in self.components.iter_mut() {
                    component.set_active(self.current_tab.clone());
                }
            },
            KeyCode::Char('h') | KeyCode::Left => {
                self.move_left();
                for component in self.components.iter_mut() {
                    component.set_active(self.current_tab.clone());
                }
            },
            _ => {
                for component in self.components.iter_mut() {
                    component.handle_key_event(key)?;
                }
            },
        }
        Ok(None)
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        if let Action::SetActiveScreen(Mode::Explorer) = action {
            self.is_active = true;
            self.current_tab = CurrentTab::Blocks;
            for component in self.components.iter_mut() {
                component.set_active(self.current_tab.clone());
            }
        }
        for component in self.components.iter_mut() {
            component.update(action.clone())?;
        }
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let screen = super::screen_layout(area);
        for component in self.components.iter_mut() {
            component.draw(frame, screen)?;
        }
        Ok(())
    }
}

pub fn explorer_layout(area: Rect) -> [Rect; 3] {
    Layout::vertical([
        Constraint::Percentage(30),
        Constraint::Percentage(40),
        Constraint::Percentage(30),
    ]).areas(area)
}

pub fn explorer_header_layout(area: Rect) -> [Rect; 2] {
    let [header, _, _] = explorer_layout(area);
    Layout::horizontal([
        Constraint::Percentage(50),
        Constraint::Percentage(50),
    ]).areas(header)
}

pub fn explorer_info_layout(area: Rect) -> [Rect; 2] {
    let [info, _]  = explorer_header_layout(area);
    Layout::vertical([
        Constraint::Percentage(100),
        Constraint::Percentage(100),
    ]).areas(info)
}

pub fn explorer_block_info_layout(area: Rect) -> [Rect; 3] {
    let [blocks, _]  = explorer_info_layout(area);
    Layout::horizontal([
        Constraint::Percentage(33),
        Constraint::Percentage(33),
        Constraint::Percentage(33),
    ]).flex(Flex::SpaceBetween).areas(blocks)
}

pub fn explorer_era_info_layout(area: Rect) -> [Rect; 2] {
    let [_, blocks]  = explorer_info_layout(area);
    Layout::horizontal([
        Constraint::Percentage(50),
        Constraint::Percentage(50),
    ]).flex(Flex::SpaceBetween).areas(blocks)
}

pub fn explorer_scrollbars_layout(area: Rect) -> [Rect; 2] {
    let [_, place, _] = explorer_layout(area);
    Layout::horizontal([
        Constraint::Percentage(50),
        Constraint::Percentage(50),
    ]).areas(place)
}
